﻿// Copyright (c) 2018 ManusVR

using ManusVR.SDK.Apollo;
using System;
using System.Runtime.InteropServices;

namespace ManusVR.SDK.Manus
{
    using manus_session_t = IntPtr;

    public enum coor_up_t
    {
        COOR_X_UP = 0,
        COOR_Y_UP,
        COOR_Z_UP
    }

    public enum coor_handed_t
    {
        COOR_LEFT_HANDED = 0,
        COOR_RIGHT_HANDED
    }

    [Serializable]
    public struct quat_t
    {
        public double x, y, z, w;

#if UNITY_EDITOR || UNITY_STANDALONE
        public static implicit operator UnityEngine.Quaternion(quat_t b)
        {
            return new UnityEngine.Quaternion((float)b.x, (float)b.y, (float)b.z, (float)b.w);
        }

        public static implicit operator quat_t(UnityEngine.Quaternion b)
        {
            quat_t a = new quat_t();
            a.x = b.x;
            a.y = b.y;
            a.z = b.z;
            a.w = b.w;
            return a;
        }
#endif
    }

    public struct vector_t
    {
        public double x, y, z;

#if UNITY_EDITOR || UNITY_STANDALONE
        public static implicit operator UnityEngine.Vector3(vector_t b)
        {
            return new UnityEngine.Vector3((float)b.x, (float)b.y, (float)b.z);
        }

        public static implicit operator vector_t(UnityEngine.Vector3 b)
        {
            vector_t a = new vector_t();
            a.x = b.x;
            a.y = b.y;
            a.z = b.z;
            return a;
        }
#endif

    }
    [Serializable]
    public struct finger_t
    {
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 5)]
        public pose_t[] joints;
    }

    [Serializable]
    public struct manus_hand_raw_t
    {
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 2)]
        public quat_t[] imu;
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 10)]
        public double[] finger_sensor;
    }

    [Serializable]
    public struct manus_hand_t
    {
        public manus_hand_raw_t raw;
        public quat_t wrist;
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 5)]
        public finger_t[] fingers;
    }

    public struct finger_profile_t
    {
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
        public double[] bones;
    }

    public struct hand_profile_t
    {
        public double wrist;
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 5)]
        public finger_profile_t[] fingers;
    }

    /*! Skeletal model of the arm which contains a quaternion for each joint and a position for the hand. */
    public struct ik_arm_t
    {
        public pose_t shoulder; // OUTPUT
        public quat_t upperArm; // OUTPUT
        public pose_t lowerArm; // INPUT, set this at the controller position
    }

    /*! Skeletal model for the upper body of the player. */
    public struct ik_body_t
    {
        public pose_t head; // INPUT, set this at the headset position
        public ik_arm_t left, right;
    }

    public struct ik_profile_t
    {
        public double shoulderLength;
        public double upperArmLength;
        public double lowerArmLength;
        public double upperNeckLength;
        public double lowerNeckLength;
        public vector_t upperNeckOffset;
        public hand_profile_t handProfile;
    }

    /*!
    *   \brief Glove class
    *
    */
    public class Manus
    {
        /*! Initialize the Manus SDK.
        *
        *   Must be called before any other function in the SDK.
        *   This function should only be called once.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusInit(out manus_session_t session);

        /*! Shutdown the Manus SDK.
        *
        *   Must be called when the SDK is no longer
        *   needed.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusExit(manus_session_t session);

        /*! Get the dongle id of every dongle in use. Returns the number of dongles in use if count is
        * not the same as the number of dongles in use.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern UInt32 ManusGetDongleIDs(manus_session_t session, UInt32[] out_ids, UInt32 count);

        /*! Set the ouput power of the vibration motor.
        *
        *   This function should only be called once every 2ms. Higher update rates will be
        *   supported in later versions of the SDK.
        *
        *  \param device The device type to vibrate.
        *  \param power The power of the vibration motor ranging from 0 to 1 (ex. 0.5 = 50% power).
        *  \param milliseconds The duration of the vibration.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusSetVibration(manus_session_t session, device_type_t device_type, float power, UInt16 milliseconds);

        /*! Set the ouput power of the vibration motor.
        *
        *   This function should only be called once every 2ms. Higher update rates will be
        *   supported in later versions of the SDK.
        *
        *  \param dongle_id The ID of the dongle that this call is meant for.
        *  \param device The device type to vibrate.
        *  \param power The power of the vibration motor ranging from 0 to 1 (ex. 0.5 = 50% power).
        *  \param milliseconds The duration of the vibration.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusSetVibration_id(manus_session_t session, UInt32 dongle_id, device_type_t device_type, float power, UInt16 milliseconds);

        /*! Switch to a different saved user profile.
        *
        *  \param index The profile index.
        *
        *  \see ManusGetProfileCount
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusSelectProfile(manus_session_t session, UInt32 index);

        /*! Get the IK user profile identifier.
        *
        *  \param index The profile index.
        *
        *  \see ManusGetProfileCount
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern string ManusGetProfileName(manus_session_t session, UInt32 index);

        /*! Creates a new user profile and returns its index.
        *
        *   The profile can be selected by passing the index to ManusSelectProfile().
        *
        *  \param name User profile identifier.
        *  \returns The index of the newly added profile.
        *
        *  \see ManusSelectProfile
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern UInt32 ManusCreateProfile(manus_session_t session, String name);

        /*! Get the total number of user profiles.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern UInt32 ManusGetProfileCount(manus_session_t session);

        /*! Get the data from the currently selected user profile.
        *
        *   This profile is used by the IK system for inverse kinematics and for the skeletal hand model.
        *   It should also be used by the game to scale the in-game model to fit the user profile.
        *
        *  \see ManusUpdateIK
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetProfile(manus_session_t session, out ik_profile_t out_profile);

        /*! Change currently selected user profile data.
        *
        *   This changes the working copy of the currently selected profile. If you want to save
        *   the current working copy to disk you should call the commit function.
        *
        *  \see ManusCommitProfile
        *
        *  \param profile The new profile data to set in the working copy.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusSetProfile(manus_session_t session, ik_profile_t profile);

        /*! Revert the working copies to the profiles saved on the file system.
        *
        *   This reverts all changes made to the working copy that hasn't been committed.
        *
        *  \param ManusCommitProfile
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusRevertProfiles(manus_session_t session);

        /*! Write the working copies of all user profiles to the file system.
        *
        *   After this call the old profile data can no longer be retrieved.
        *
        *  \param ManusRevertProfile
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusCommitProfiles(manus_session_t session);

        /*! Calculate the shoulder and elbow position.
        *
        *   This calculates shoulder and elbow position based on the wrist and head position.
        *
        *  \param model The model of the body with both the inputs and the outputs of this calculation.
        *  \param settings The settings of the person used for the calculations, such as arm lengths.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusUpdateIK(manus_session_t session, ref ik_body_t inout_model);

        /*! Check if the device is connected.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern bool ManusIsConnected(manus_session_t session, device_type_t device);
        /*! Check if the device is connected.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern bool ManusIsConnected_id(manus_session_t session, UInt32 dongle_id, device_type_t device);

        /*! Sets the coordinate system for all orientations and vectors returned from the SDK.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusSetCoordinateSystem(manus_session_t session, coor_up_t up, coor_handed_t handed);

        /*! Get the current battery level in percent.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetBatteryLevel(manus_session_t session, device_type_t device_type, out byte out_percentage);
        /*! Get the current battery level in percent.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetBatteryLevel_id(manus_session_t session, UInt32 dongle_id, device_type_t device_type, out byte out_percentage);

        /*! Get the current battery level in millivolt.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetBatteryVoltage(manus_session_t session, device_type_t device_type, out UInt16 out_millivolts);
        /*! Get the current battery level in millivolt.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetBatteryVoltage_id(manus_session_t session, UInt32 dongle_id, device_type_t device_type, out UInt16 out_millivolts);

        /*! Get the signal strength.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetSignalStrength(manus_session_t session, device_type_t device_type, out UInt16 out_rssi);
        /*! Get the signal strength.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetSignalStrength_id(manus_session_t session, UInt32 dongle_id, device_type_t device_type, out UInt16 out_rssi);

        /*! Get the data from the glove and convert it to a skeletal model. The skeletal model is scaled
        *   according to the user profile.
        *
        *  \see ManusGetProfile
        *
        *  \param device_type The device type to query.
        *  \param out_hand The output struct, also contains the raw data from the device.
        *  \param timeout The time in milliseconds to block waiting on a new packet.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetHand(manus_session_t session, device_type_t device_type, out manus_hand_t out_hand);
        /*! Get the data from the glove and convert it to a skeletal model. The skeletal model is scaled
        *   according to the user profile.
        *
        *  \see ManusGetProfile
        *
        *  \param dongle_id The ID of the dongle that this call is meant for.
        *  \param device_type The device type to query.
        *  \param out_hand The output struct, also contains the raw data from the device.
        *  \param timeout The time in milliseconds to block waiting on a new packet.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetHand_id(manus_session_t session, UInt32 dongle_id, device_type_t device_type, out manus_hand_t out_hand);

        /*! Sets the dongle channel
        *
        *  \param channel The channel to set the dongle to.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusSetChannel(manus_session_t session, UInt32 channel);
        /*! Sets the dongle channel
        *
        *  \param dongle_id The ID of the dongle that this call is meant for.
        *  \param channel The channel to set the dongle to.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusSetChannel_id(manus_session_t session, UInt32 dongle_id, UInt32 channel);

        /*! Get the raw data from the glove, this does not include any of the skeletal model data. If you're also
        *   interested in the skeletal model you should call ManusGetHand() instead.
        *
        *  \see ManusGetHand
        *
        *  \param device_type The device type to query.
        *  \param out_hand The output struct.
        *  \param timeout The time in milliseconds to block waiting on a new packet.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetHandRaw(manus_session_t session, device_type_t device_type, out manus_hand_raw_t out_hand);
        /*! Get the raw data from the glove, this does not include any of the skeletal model data. If you're also
        *   interested in the skeletal model you should call ManusGetHand() instead.
        *
        *  \see ManusGetHand
        *
        *  \param dongle_id The ID of the dongle that this call is meant for.
        *  \param device_type The device type to query.
        *  \param out_hand The output struct.
        *  \param timeout The time in milliseconds to block waiting on a new packet.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusGetHandRaw_id(manus_session_t session, device_type_t device_type, out manus_hand_raw_t out_hand);

        /*! Gets a string that describes the error code. This can be used to display error messages
        *   in the user interface.
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern string ManusErrorString(UInt32 error);

        /*! Adds a virtual dongle that emulates two gloves connected to the API.
        *
        *  \see ManusRemoveDebugDevice
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusAddDebugDevice(manus_session_t session);

        /*! Removes the virtual dongle.
        *
        *  \see ManusAddDebugDevice
        */
        [DllImport("Manus", CallingConvention = CallingConvention.Cdecl)]
        public static extern manus_ret_t ManusRemoveDebugDevice(manus_session_t session);
    }
}
